 def _start
 def _init_data

# Data initialization structure
#
# This structure will be filled in by the ELF conversion program
# after link time.
_init_data:
  data   # Start of data section
  data   # Size of data section
  data   # Location of initial contents of data section
  data   # Start of BSS section
  data   # Size of BSS section

# Entry point for C runtime initilization code
#
# This will set initial values for data memory and otherwise create
# the conditions needed for the C environment
_start:

# Disable interrupts
#
# This will disable console reset via keyboard, music playback,
# sprite automotion, etc. But will make the environment much simpler
# to work in. Users can re-enable interrupts later if they wish.
  limi 0

# Set initial workspace
#
# The registers will be located at the start of scratchpad memory 
  lwpi >8300

# Initialize data segment to known values
  li   r0, _init_data  # Point to data initialization structure
  mov  *r0+, r1        # R1 = Start of data section
  mov  *r0+, r2        # R2 = Size of data section
  mov  *r0+, r3        # R3 = Location of initial data
  
  mov  *r3+, *r1+      # Copy loop for data initialization
  dect r2
  jgt  $-4

# Initialize BSS segment to zeroes
  mov  *r0+, r1        # R1 = Start of BSS section
  mov  *r0, r2         # R2 = Size of BSS section
  
  clr  *r1+            # Clearing loop for BSS initialization
  dect r2
  jgt  $-4

# Create the stack
  li sp, >4000

# Start running C code
  bl @main

# We cannot properly handle a return from C code.
# Instead, just spin here until the console is reset
  jmp $

