#define VDP_READ_DATA_REG	(*(volatile char*)0x8800)
#define VDP_WRITE_DATA_REG	(*(volatile char*)0x8C00)
#define VDP_ADDRESS_REG		(*(volatile char*)0x8C02)
#define VDP_READ_FLAG		0x00
#define VDP_WRITE_FLAG		0x40
#define VDP_REG_FLAG		0x80
#define SEED			(*(int*)0x83C0)
#define STATUS			(*(char*)0x837C)
#define KEY_UNIT		(*(char*)0x8374)
#define KEY			(*(char*)0x8375)
#define JOYST_X			(*(char*)0x8376)
#define JOYST_Y			(*(char*)0x8377)
#define FAC			(*(volatile int*)0x834A)
#define PAB_NAMELENGTH_ADDR	(*(volatile int*)0x8356)
#define INPUT			0x4
#define OUTPUT			0x2
#define SOUND_PLAYING		(*(char*)0x83CE)
#define SOUND_LIST_POINTER	(*(int*)0x83CC)
#define GPL_R14_LOW		(*(char*)0x83FD)

#define TRANSPARENT		0
#define BLACK			1
#define MEDIUM_GREEN		2
#define LIGHT_GREEN		3
#define	DARK_BLUE		4
#define LIGHT_BLUE		5
#define DARK_RED		6
#define CYAN			7
#define MEDIUM_RED		8
#define LIGHT_RED		9
#define DARK_YELLOW		10
#define LIGHT_YELLOW		11
#define DARK_GREEN		12
#define MAGENTA			13
#define GRAY			14
#define WHITE			15

int EOF;

void vdp_copy_from_sys(int index, char* src, int size) {
	char* end = src + size;
	VDP_ADDRESS_REG = index;
	VDP_ADDRESS_REG = (char)(index >> 8) | VDP_WRITE_FLAG;
	while(src < end) VDP_WRITE_DATA_REG = *src++;
}

void vdp_write_reg(int i,char c) {
	VDP_ADDRESS_REG=c;
	VDP_ADDRESS_REG=i|VDP_REG_FLAG;
}

void interrupts() {
	asm("limi 2");
	asm("limi 0");
}

char random_byte(char c) {
	SEED=SEED*0x6FE5+0x7AB9;
	int i=((unsigned)SEED>>8)+(SEED<<8);
	return (unsigned)i%c;
}

char vdp_read_byte(int index) {
	VDP_ADDRESS_REG = index;
	VDP_ADDRESS_REG = (char)(index >> 8) | VDP_READ_FLAG;
	asm("nop");
	return VDP_READ_DATA_REG;
}

void vdp_read_bytes(int i,char* buf,int n) {
	char* end=buf+n;
	VDP_ADDRESS_REG=i;
	VDP_ADDRESS_REG=(char)(i>>8)|VDP_READ_FLAG;	
	while(buf<end)*buf++=VDP_READ_DATA_REG;
}

int key_scan(char* key) {
	KEY_UNIT=5;
	asm("bl @kscan");
	*key=KEY;
	char c=STATUS&0x20;
	return c;
}

void gpl_link(int addr) {
	asm(
		"mov %0,r1\n\t"
		"blwp @gpllnk"
		:
		:"r"(addr)
		:"r1"
	);
}

int dsr_link(int op) {
	volatile int stat;
	asm(
		"mov %1,r1\n\t"
		"blwp @dsrlnk\n\t"
		"stst %0"
		:"=r"(stat)
		:"r"(op)
		:"r0","r1"
	);
	return stat;
}

typedef struct {
	char opcode;
	volatile char flags;
	int buffer_address;
	char record_length;
	char character_count;
	int record_number;
	char screen_offset;
	char name_length;
	char name[50];
} pab;

void close(pab* p);

int file_operation(pab* p,int op) {
	p->opcode=op;
	EOF=0;
	p->flags=p->flags&0x1F;
	vdp_copy_from_sys(0xF80,(char*)p,sizeof(pab));
	PAB_NAMELENGTH_ADDR=0xF89;
	int stat=dsr_link(8);
	vdp_read_bytes(0xF80,(char*)p,sizeof(pab));
	int err=p->flags>>5;
	if(err==5) {
		EOF=1;
		err=0;
	}
	else {
		if(err)err++;
		else if(stat&0x2000)err=1;
	}
	
	if(err && p->opcode!=1)close(p);
	return err;	
}

char *strcpy(char *dest,const char *src) {
   char *save=dest;
   while(*dest++=*src++);
   return save;
}

int strlen(const char* str) {
    const char *s;
    for (s=str;*s;++s);
    return(s-str);
}

int open(pab* p,const char* s,char mode) {
	p->flags=0x10|mode;
	p->buffer_address=0x1000;
	p->record_length=80;
	strcpy(p->name,s);
	p->name_length=strlen(p->name);
	return file_operation(p,0);
}

void close(pab* p) {	
	file_operation(p,1);
}

int write(pab* p,char* buf,char n) {
	vdp_copy_from_sys(0x1000,buf,n);
	p->character_count=n;
	return file_operation(p,3);
}

int read(pab* p,char* buf) {
	int ret=file_operation(p,2);
	vdp_read_bytes(0x1000,buf,p->character_count);
	return ret;
}

void defchar(int i,int pat1,int pat2,int pat3,int pat4) {
	vdp_copy_from_sys(i*8+0x800,(char*)&pat1,2);
	vdp_copy_from_sys(i*8+0x800+2,(char*)&pat2,2);
	vdp_copy_from_sys(i*8+0x800+4,(char*)&pat3,2);
	vdp_copy_from_sys(i*8+0x800+6,(char*)&pat4,2);
}

void color(int charset,char fg,char bg) {
	char c=bg+(fg<<4);
	vdp_copy_from_sys(charset+0x380,&c,1);
}

void position(int i,char y,char x) {
	int i2=x+(y<<8);
	vdp_copy_from_sys(0x300+i*4,(char*)&i2,2);
}

void sprite(int i,char character,char color) {
	int i2=color+(character+128<<8);
	vdp_copy_from_sys(0x300+i*4+2,(char*)&i2,2);
}

void defsprite(int i,int pat1,int pat2,int pat3,int pat4) {
	vdp_copy_from_sys(i*8+0x400,(char*)&pat1,2);
	vdp_copy_from_sys(i*8+0x400+2,(char*)&pat2,2);
	vdp_copy_from_sys(i*8+0x400+4,(char*)&pat3,2);
	vdp_copy_from_sys(i*8+0x400+6,(char*)&pat4,2);
}

void defsprite2(int i,int* pat) {
	vdp_copy_from_sys(i*8+0x400,(char*)pat,8);
}

void magnify(int i) {
	vdp_write_reg(1,0xE0|(i-1));
}

void defchar2(int i,int* pat) {
	vdp_copy_from_sys(i*8+0x800,(char*)pat,8);
}

void load_standard_characters() {
	FAC=0x900;
	gpl_link(0x18);
	FAC=0xB00;
	gpl_link(0x4A);
}

void screen(char color) {
	vdp_write_reg(7,color);
}

void pchar(char c,int y,int x) {
	vdp_copy_from_sys(y*32+x,&c,1);
}

void hchar(char c,int y,int x,int rep) {
	int i;
	for(i=0;i<rep;i++)vdp_copy_from_sys(y*32+x+i,&c,1);
}

void vchar(char c,int y,int x,int rep) {
	int i;
	for(i=0;i<rep;i++)vdp_copy_from_sys((y+i)*32+x,&c,1);
}

void reverse(char s[]) {
	int i,j;
	char c;
	for(i=0,j=strlen(s)-1;i<j;i++,j--) {
	 c = s[i];
	 s[i] = s[j];
	 s[j] = c;
	}
}

void itoa(int n,char s[]) {
	int i,sign;
	if((sign=n)<0)n=-n;
	i=0;
	do { s[i++]=n%10+'0'; } while((n/=10)>0);
	if(sign<0)s[i++]='-';
	s[i]='\0';
	reverse(s);
}

void display_at(int y,int x,char* c) {
	vdp_copy_from_sys(y*32+x,c,strlen(c));
}

void sound(char* c,int n) {
	vdp_copy_from_sys(0x1800,c,n);
	SOUND_LIST_POINTER=0x1800;
	GPL_R14_LOW=GPL_R14_LOW|1;
	SOUND_PLAYING=1;
}

void clear() {
	int i=0;
	VDP_ADDRESS_REG=i;
	VDP_ADDRESS_REG=(char)(i>>8)|VDP_WRITE_FLAG;
	for(i=0;i<32*24;i++)VDP_WRITE_DATA_REG=' ';
}

//******************************************* RUSH HOUR

#define TOP 0
#define MIDDLE_VERTICAL 1
#define BOTTOM 2
#define LEFT 3
#define MIDDLE_HORIZONTAL 4
#define RIGHT 5

#define UP 0
#define DOWN 2

#define RED_BLOCK_CHARS 128
#define YELLOW_BLOCKS_CHARS 144
#define SELECTED_BLOCK_CHARS 160
#define WALL_CHARS 176

int block_patterns[9][4]= {
	{ 0x001F, 0x3F7F, 0x7F7F, 0x7F7F }, // TL
	{ 0x00FF, 0xFFFF, 0xFFFF, 0xFFFF }, // TOP
	{ 0x00F8, 0xFCFE, 0xFEFE, 0xFEFE }, // TR
	{ 0x7F7F, 0x7F7F, 0x7F7F, 0x7F7F }, // ML
	{ 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF }, // M
	{ 0xFEFE, 0xFEFE, 0xFEFE, 0xFEFE }, // MR
	{ 0x7F7F, 0x7F7F, 0x7F3F, 0x1F00 }, // BL
	{ 0xFFFF, 0xFFFF, 0xFFFF, 0xFF00 }, // BOT
	{ 0xFEFE, 0xFEFE, 0xFEFC, 0xF800 }  // BR
};

int wall_patterns[8][4]= {
	{ 0xFF80, 0x8080, 0x8080, 0x8081 }, // TL
	{ 0xFF00, 0x0000, 0x0000, 0x00FF }, // HORIZ
	{ 0xFF01, 0x0101, 0x0101, 0x0181 }, // TR
	{ 0x8181, 0x8181, 0x8181, 0x8181 }, // VERT
	{ 0x8181, 0x8181, 0x8181, 0x81FF }, // BOT
	{ 0xFF81, 0x8181, 0x8181, 0x8181 }, // TOP
	{ 0x8101, 0x0101, 0x0101, 0x01FF }, // BR
	{ 0x8180, 0x8080, 0x8080, 0x80FF }  // BL
};

int sprite_patterns[8][4]= {
	{ 0x1814, 0x1211, 0x1010, 0x1010 }, // BLACK
	{ 0x1010, 0x1215, 0x1900, 0x0000 },
	{ 0x0000, 0x0000, 0x8040, 0x2010 },
	{ 0x0870, 0x4020, 0x2090, 0x9060 },
	{ 0x0008, 0x0C0E, 0x0F0F, 0x0F0F }, // WHITE
	{ 0x0F0F, 0x0D08, 0x0000, 0x0000 },
	{ 0x0000, 0x0000, 0x0080, 0xC0E0 },
	{ 0xF080, 0x80C0, 0xC060, 0x6000 }
};

char block_parts[6][9]= {
	{ 0,1,2,3,4,5,3,4,5 }, //TOP
	{ 3,4,5,3,4,5,3,4,5 }, //MIDDLE_VERTICAL
	{ 3,4,5,3,4,5,6,7,8 }, //BOTTOM
	{ 0,1,1,3,4,4,6,7,7 }, //LEFT
	{ 1,1,1,4,4,4,7,7,7 }, //MIDDLE_HORIZONTAL
	{ 1,1,2,4,4,5,7,7,8 }  //RIGHT
};

char victory_sound[32]= {
	9, 0x85, 0x0D, 0xA9, 0x0A, 0xCE, 0x08, 0x97, 0xB7, 0xD7, 4,
	3, 0x9F, 0xBF, 0xDF, 4,
	9, 0x85, 0x0D, 0xA9, 0x0A, 0xCE, 0x08, 0x97, 0xB7, 0xD7, 48,
	3, 0x9F, 0xBF, 0xDF, 0
};

typedef struct {
	char x;
	char y;
} point;

point cursor= { 3, 3 };
char blocks[6][6];
#define LEVELS_NUMBER 32
const char levels[LEVELS_NUMBER][6][6];
int selected_block=0;
int level;
int score;
int current_move;
int scores[LEVELS_NUMBER];
int finished;
int first_move=1;

void show_parts(int type,char color,int x,int y) {
	int i=0;
	int x2,y2;
	char c;
	for(y2=y;y2<=y+2;y2++) {
		for(x2=x;x2<=x+2;x2++) {
			c=block_parts[type][i]+color;
			vdp_copy_from_sys((y2+3)*32+x2+7,&c,1);
			i++;
		}
	}
}

void show_space(int x,int y) {
	int x2=x;
	int y2=y;
	char c;
	do {
		do {
			c=' ';
			vdp_copy_from_sys((y2+3)*32+x2+7,&c,1);
			x2++;
		} while(x2<=x+2);
		x2=x;
		y2++;
	} while(y2<=y+2);	
}

void show_blocks() {
	char c,c2,c3;
	int x,y,x2,y2;
	for(x=0;x<=5;x++) {
		for(y=0;y<=5;y++) {
			c=blocks[y][x];
			if(c) {
				c3=0;
				
				y2=y;
				x2=x-1;
				if(x2>=0)c2=blocks[y2][x2];
				else c2=0;
				if(c2==c)c3=RIGHT;			
				x2=x+1;
				if(x2<=5)c2=blocks[y2][x2];
				else c2=0;
				if(c2==c) {
					if(!c3)c3=LEFT;
					else c3=MIDDLE_HORIZONTAL;
				}
				
				x2=x;			
				y2=y-1;
				if(y2>=0)c2=blocks[y2][x2];
				else c2=0;
				if(c2==c)c3=BOTTOM;
				y2=y+1;
				if(y2<=5)c2=blocks[y2][x2];
				else c2=0;
				if(c2==c) {
					if(!c3)c3=TOP;
					else c3=MIDDLE_VERTICAL;
				}
				
				if(c==selected_block)c2=SELECTED_BLOCK_CHARS;
				else {
					if(c==1)c2=RED_BLOCK_CHARS;
					else c2=YELLOW_BLOCKS_CHARS;
				}
				show_parts(c3,c2,x*3,y*3);
			}
			else show_space(x*3,y*3);
		}
	}
}

void show_wall() {
	pchar(WALL_CHARS+0,2,6);
	hchar(WALL_CHARS+1,2,7,18);
	pchar(WALL_CHARS+2,2,25);
	vchar(WALL_CHARS+3,3,25,5);
	pchar(WALL_CHARS+4,8,25);
	pchar(WALL_CHARS+5,12,25);
	vchar(WALL_CHARS+3,13,25,8);
	pchar(WALL_CHARS+6,21,25);
	hchar(WALL_CHARS+1,21,7,18);
	pchar(WALL_CHARS+7,21,6);
	vchar(WALL_CHARS+3,3,6,18);
}

void show_cursor() {
	position(0,cursor.y*8*3+4*8,cursor.x*8*3+8*8-1);
	position(1,cursor.y*8*3+4*8,cursor.x*8*3+8*8-1);
}

void show_score() {
	char s[7];
	if(score) {
		itoa(score,s);		
		int n=strlen(s);
		char* s2="MOVES ";
		display_at(0,25-n,s2);
		display_at(0,31-n,s);
	}
	if(scores[level]) {
		char* s2="RECORD";
		display_at(23,1,s2);
		itoa(scores[level],s);		
		display_at(23,8,s);
	}
}

void selection() {
	if(selected_block==0) {
		if(!finished) {
			selected_block=blocks[cursor.y][cursor.x];
			current_move=0;
			show_blocks();
		}
	}
	else {
		selected_block=0;
		if(current_move) {
			score++;
			show_score();
		}
		show_blocks();
	}
}

void endgame() {
	sound(victory_sound,32);
	show_space(4*3,2*3);
	show_parts(LEFT,RED_BLOCK_CHARS,5*3,2*3);
	show_parts(RIGHT,RED_BLOCK_CHARS,6*3,2*3);
	selected_block=0;
	finished=1;
	score++;
	if(scores[level]==0 || scores[level]>score)scores[level]=score;
	show_score();
	pab p;
	if(!open(&p,"DSK1.RHSC",OUTPUT)) {
		write(&p,(char*)&level,2);
		write(&p,(char*)scores,LEVELS_NUMBER*2);		
		close(&p);
	}
}

int move_block(char dx,char dy) {
	int x=cursor.x; int y=cursor.y;
	char c=selected_block;
	int ret=0;
	x+=dx; y+=dy;
	if(blocks[y][x]==c || blocks[y][x]==0 || x==6 && y==2) {
		int xfront=x;
		int yfront=y;
		int xtail=cursor.x;
		int ytail=cursor.y;
		
		if(dx) {
			while(blocks[yfront][xfront]==c && xfront<6 && xfront>0)xfront+=dx;
			while(blocks[ytail][xtail]==c && xtail<6 && xtail>0)xtail-=dx;
		}
		else {
			while(blocks[yfront][xfront]==c && yfront<5 && yfront>0)yfront+=dy;
			while(blocks[ytail][xtail]==c && ytail<5 && ytail>0)ytail-=dy;
		}
		if(blocks[ytail][xtail]!=c) { xtail+=dx; ytail+=dy; }
		if(dx && abs(xtail-xfront)>1 || dy && abs(ytail-yfront)>1) {
			if(xfront==6 && yfront==2 && c==1) {
				endgame();
				ret=1;
			}
			else {
				if(xfront==6)xfront=5;
				if(blocks[yfront][xfront]==0) {
					blocks[yfront][xfront]=c;
					blocks[ytail][xtail]=0;
					ret=1;
				}
			}
		}
	}
	if(ret && !finished) {
		show_blocks();
		current_move+=dx+dy;
	}
	return ret;
}

void move(int dx,int dy) {
	int ok=1;
	if(selected_block)ok=move_block(dx,dy);
	if(ok) {
		cursor.x+=dx;
		if(cursor.x<0)cursor.x=0;
		if(cursor.x>5)cursor.x=5;
		cursor.y+=dy;
		if(cursor.y<0)cursor.y=0;
		if(cursor.y>5)cursor.y=5;
		show_cursor();
	}
	if(first_move) {
		first_move=0;
		hchar(' ',23,22,9);
	}
}

void show_level() {
	hchar(' ',0,6,32);
	if(scores[level])hchar(' ',23,7,25);
	else hchar(' ',23,0,32);
	char* c="LEVEL";
	display_at(0,1,c);
	char s[7];
	itoa(level+1,s);
	display_at(0,7,s);
	show_score();
	show_space(6*3,2*3);
	show_blocks();
}

void load_level() {
	int x,y;
	for(x=0;x<=5;x++) {
		for(y=0;y<=5;y++)blocks[y][x]=levels[level][y][x];
	}
	score=0;
	selected_block=0;
	finished=0;
	show_level();
}

void aid() {
	position(0,208,0);
	clear();
	char* s1="----- RUSH HOUR -----";
	display_at(1,5,s1);
	char* s2="ESDX TO MOVE";
	display_at(5,1,s2);
	char* s3="ENTER/SPACE TO SELECT A BLOCK";
	display_at(7,1,s3);
	char* s4="N TO LOAD NEXT LEVEL";
	display_at(9,1,s4);
	char* s5="P TO LOAD PREVIOUS LEVEL";
	display_at(11,1,s5);
	char* s6="LUCIEN BAUMANN 2011";
	display_at(22,12,s6);
	char c;
	while(!key_scan(&c));
	clear();
	magnify(3);
	show_cursor();
	show_wall();
	show_level();
}

char string[10];
void main() {
	load_standard_characters();
	screen(CYAN);
	int i;
	for(i=4;i<=11;i++)color(i,BLACK,CYAN);
	color(16,DARK_RED,CYAN);
	color(17,DARK_RED,CYAN);
	for(i=0;i<=8;i++)defchar2(RED_BLOCK_CHARS+i,block_patterns[i]);
	color(18,DARK_BLUE,CYAN);
	color(19,DARK_BLUE,CYAN);
	for(i=0;i<=8;i++)defchar2(YELLOW_BLOCKS_CHARS+i,block_patterns[i]);
	color(20,MEDIUM_GREEN,CYAN);
	color(21,MEDIUM_GREEN,CYAN);
	for(i=0;i<=8;i++)defchar2(SELECTED_BLOCK_CHARS+i,block_patterns[i]);	
	color(22,DARK_YELLOW,LIGHT_YELLOW);
	for(i=0;i<=7;i++)defchar2(WALL_CHARS+i,wall_patterns[i]);

	for(i=0;i<=7;i++)defsprite2(i,sprite_patterns[i]);
	sprite(0,0,BLACK);
	sprite(1,4,WHITE);
	magnify(3);
	
	show_cursor();
	show_wall();

	pab p;
	int err=1;
	if(!open(&p,"DSK1.RHSC",INPUT)) {
		read(&p,(char*)&level);
		err=read(&p,(char*)scores);
		close(&p);
	}
	
	if(err) {
		level=0;
		int i;
		for(i=0;i<LEVELS_NUMBER;i++)scores[i]=0;
	}
	load_level();
	char* s="PRESS AID";
	display_at(23,22,s);

	char c;
	while(1) {
		interrupts();
		if(key_scan(&c)) {
			magnify(3);
			
			switch(c) {
				case 'S':
				case 8:
					move(-1,0);
					break;
				case 9:
				case 'D':
					move(1,0);
					break;
				case 10:
				case 'X':
					move(0,1);
					break;
				case 11:
				case 'E':
					move(0,-1);
					break;
				case 13:
				case ' ':
					selection();
					break;
				case 'N':
					level++;
					if(level>=LEVELS_NUMBER)level=0;
					load_level();
					break;
				case 'P':
					level--;
					if(level<0)level=LEVELS_NUMBER-1;
					load_level();
					break;
				case  1:
					aid();
					break;
			}
		}
	}
}

const char levels[LEVELS_NUMBER][6][6]= {
	{{  2,  2,  2,  0,  0,  3 }, //Beginner 1
	 {  0,  0,  4,  0,  0,  3 },
	 {  1,  1,  4,  0,  0,  3 },
	 {  5,  0,  4,  0,  6,  6 },
	 {  5,  0,  0,  0,  7,  0 },
	 {  8,  8,  8,  0,  7,  0 }},
	{{  0,  0,  0,  0,  0,  0 }, //2
	 {  0,  0,  2,  3,  4,  0 },
	 {  1,  1,  2,  3,  4,  0 },
	 {  5,  5,  6,  3,  4,  0 },
	 {  0,  7,  6,  0,  0,  0 },
	 {  0,  7,  8,  8,  0,  0 }},
	{{  2,  3,  3,  4,  4,  0 }, //3
	 {  2,  0,  5,  0,  0,  0 },
	 {  1,  1,  5,  6,  7,  0 },
	 {  0,  0,  0,  6,  7,  0 },
	 {  8,  8,  8,  6,  7,  0 },
	 {  0,  0,  0,  0,  0,  0 }},
	{{  0,  0,  0,  0,  0,  0 }, //4
	 {  0,  0,  0,  2,  3,  4 },
	 {  1,  1,  0,  2,  3,  4 },
	 {  5,  5,  0,  2,  3,  0 },
	 {  0,  6,  7,  7,  0,  0 },
	 {  0,  6,  8,  8,  0,  0 }},
	{{  0,  0,  2,  0,  3,  3 }, //5
	 {  0,  0,  2,  0,  0,  4 },
	 {  1,  1,  2,  0,  0,  4 },
	 {  0,  5,  6,  6,  6,  4 },
	 {  0,  5,  7,  7,  8,  0 },
	 {  0,  0,  0,  0,  8,  0 }},
	{{  0,  0,  2,  3,  4,  4 }, //6
	 {  0,  0,  2,  3,  5,  0 },
	 {  1,  1,  2,  3,  5,  0 },
	 {  0,  6,  0,  0,  7,  0 },
	 {  0,  6,  8,  8,  7,  0 },
	 {  0,  0,  0,  0,  0,  0 }},
	{{  2,  0,  0,  3,  0,  0 }, //7
	 {  2,  0,  0,  3,  0,  0 },
	 {  1,  1,  0,  4,  0,  5 },
	 {  6,  6,  6,  4,  0,  5 },
	 {  0,  7,  8,  8,  0,  5 },
	 {  0,  7,  9,  9,  0,  0 }},
	{{  0,  0,  0,  0,  0,  0 }, //8
	 {  0,  2,  2,  2,  0,  3 },
	 {  1,  1,  4,  5,  0,  3 },
	 {  0,  0,  4,  5,  6,  6 },
	 {  7,  7,  7,  5,  0,  8 },
	 {  0,  0,  0,  0,  0,  8 }},
	{{  0,  2,  3,  4,  5,  5 }, //Intermediate 1
	 {  0,  2,  3,  4,  6,  6 },
	 {  1,  1,  3,  0,  0,  0 },
	 {  7,  8,  8,  8,  0,  9 },
	 {  7,  0,  0,  0,  0,  9 },
	 { 10, 10, 11, 11,  0,  9 }},
	{{  0,  2,  3,  3,  3,  4 }, //2
	 {  0,  2,  5,  5,  5,  4 },
	 {  1,  1,  6,  0,  7,  4 },
	 {  8,  8,  6,  9,  7,  0 },
	 { 10,  0,  6,  9,  0,  0 },
	 { 10, 11, 11,  9,  0,  0 }},
	{{  2,  2,  2,  3,  4,  5 }, //3
	 {  6,  6,  0,  3,  4,  5 },
	 {  7,  1,  1,  3,  4,  8 },
	 {  7,  0,  9,  9,  0,  8 },
	 {  0,  0, 10, 11, 11, 11 },
	 {  0,  0, 10, 12, 12, 12 }},
	{{  0,  0,  2,  3,  3,  3 }, //4
	 {  4,  0,  2,  5,  6,  6 },
	 {  4,  1,  1,  5,  7,  0 },
	 {  0,  8,  0,  0,  7,  9 },
	 {  0,  8, 10, 10,  7,  9 },
	 {  0, 11, 11, 12, 12,  0 }},
	{{  2,  3,  3,  3,  0,  4 }, //5
	 {  2,  5,  5,  5,  0,  4 },
	 {  1,  1,  0,  0,  0,  4 },
	 {  0,  6,  6,  7,  0,  0 },
	 {  0,  0,  8,  7,  9,  9 },
	 { 10, 10,  8,  7,  0,  0 }},
	{{  0,  2,  3,  3,  3,  4 }, //6
	 {  0,  2,  5,  0,  0,  4 },
	 {  1,  1,  5,  0,  6,  7 },
	 {  8,  9,  9,  9,  6,  7 },
	 {  8,  0, 10, 11, 12, 12 },
	 {  0,  0, 10, 11, 13, 13 }},
	{{  0,  2,  2,  0,  3,  3 }, //7
	 {  4,  4,  4,  0,  5,  6 },
	 {  7,  8,  1,  1,  5,  6 },
	 {  7,  8,  0,  0,  5,  9 },
	 {  7,  0, 10, 11, 11,  9 },
	 { 12, 12, 10,  0, 13, 13 }},
	{{  0,  0,  2,  2,  2,  3 }, //8
	 {  4,  4,  5,  5,  6,  3 },
	 {  7,  1,  1,  0,  6,  8 },
	 {  7,  9,  9,  0,  6,  8 },
	 {  0,  0, 10, 11, 11,  0 },
	 { 12, 12, 10,  0,  0,  0 }},
	{{  0,  2,  3,  3,  4,  0 }, //Advanced 1
	 {  0,  2,  5,  5,  4,  0 },
	 {  1,  1,  6,  7,  0,  0 },
	 {  8,  0,  6,  7,  0,  0 },
	 {  8,  0,  9,  7, 10, 10 },
	 {  8,  0,  9, 11, 11, 11 }},
	{{  0,  2,  3,  3,  3,  4 }, //2
	 {  5,  2,  0,  0,  6,  4 },
	 {  5,  2,  1,  1,  6,  0 },
	 {  7,  7,  8,  0,  6,  0 },
	 {  0,  0,  8,  9,  9,  9 },
	 {  0, 10, 10, 11, 11,  0 }},
	{{  0,  2,  3,  3,  4,  5 }, //3
	 {  0,  2,  6,  7,  4,  5 },
	 {  1,  1,  6,  7,  4,  8 },
	 {  0,  9,  9, 10,  0,  8 },
	 {  0,  0,  0, 10,  0,  0 },
	 {  0, 11, 11, 12, 12,  0 }},
	{{  0,  2,  2,  3,  4,  5 }, //4
	 {  0,  6,  7,  3,  4,  5 },
	 {  0,  6,  7,  1,  1,  5 },
	 {  8,  8,  9,  0,  0,  0 },
	 { 10,  0,  9, 11, 12, 12 },
	 { 10, 13, 13, 11,  0,  0 }},
	{{  2,  0,  0,  3,  3,  3 }, //5
	 {  2,  0,  4,  4,  5,  5 },
	 {  1,  1,  6,  0,  0,  7 },
	 {  0,  0,  6,  8,  0,  7 },
	 {  9,  9,  9,  8,  0, 10 },
	 { 11, 11, 11,  0,  0, 10 }},
	{{  2,  2,  2,  3,  0,  4 }, //6
	 {  5,  5,  6,  3,  0,  4 },
	 {  1,  1,  6,  0,  0,  7 },
	 {  8,  9,  9, 10, 10,  7 },
	 {  8,  0, 11, 11, 12, 12 },
	 { 13, 13, 13,  0,  0,  0 }},
	{{  2,  3,  4,  4,  4,  5 }, //7
	 {  2,  3,  0,  0,  6,  5 },
	 {  2,  3,  1,  1,  6,  0 },
	 {  7,  7,  8,  0,  6,  0 },
	 {  0,  0,  8,  9,  9,  9 },
	 {  0, 10, 10, 11, 11,  0 }},
	{{  0,  2,  2,  0,  3,  0 }, //8
	 {  4,  4,  4,  0,  3,  0 },
	 {  1,  1,  5,  0,  3,  6 },
	 {  7,  8,  5,  9,  9,  6 },
	 {  7,  8, 10, 10,  0, 11 },
	 { 12, 12, 12,  0,  0, 11 }},
	{{  2,  0,  3,  4,  4,  4 }, //Expert 1
	 {  2,  0,  3,  5,  0,  0 },
	 {  2,  1,  1,  5,  0,  6 },
	 {  7,  7,  8,  8,  0,  6 },
	 {  9,  9, 10, 10,  0,  6 },
	 {  0,  0,  0,  0,  0,  0 }},
	{{  2,  0,  3,  3,  4,  5 }, //2
	 {  2,  6,  6,  6,  4,  5 },
	 {  1,  1,  7,  8,  9,  0 },
	 { 10, 10,  7,  8,  9,  0 },
	 {  0,  0,  0, 11, 12, 12 },
	 {  0,  0,  0, 11, 13, 13 }},
	{{  2,  2,  0,  3,  4,  0 }, //3
	 {  5,  5,  5,  3,  4,  0 },
	 {  1,  1,  0,  3,  6,  0 },
	 {  7,  0,  8,  8,  6,  0 },
	 {  7,  0,  9, 10, 10,  0 },
	 { 11, 11,  9,  0,  0,  0 }},
	{{  0,  0,  0,  2,  0,  0 }, //4
	 {  0,  0,  3,  2,  4,  4 },
	 {  1,  1,  3,  5,  6,  7 },
	 {  8,  8,  0,  5,  6,  7 },
	 {  9, 10, 10, 10, 11,  0 },
	 {  9, 12, 12, 12, 11,  0 }},
	{{  2,  0,  3,  3,  4,  4 }, //5
	 {  2,  0,  5,  6,  6,  6 },
	 {  1,  1,  5,  0,  7,  8 },
	 {  0,  9,  9, 10,  7,  8 },
	 { 11, 11, 11, 10,  7, 12 },
	 {  0,  0,  0,  0,  0, 12 }},
	{{  2,  2,  2,  3,  4,  0 }, //6
	 {  5,  6,  6,  3,  4,  0 },
	 {  5,  1,  1,  7,  0,  0 },
	 {  5,  0,  8,  7,  9,  9 },
	 {  0,  0,  8, 10, 10, 11 },
	 { 12, 12, 12,  0,  0, 11 }},
	{{  0,  2,  3,  3,  4,  4 }, //7
	 {  0,  2,  0,  5,  6,  6 },
	 {  1,  1,  0,  5,  7,  0 },
	 {  8,  9,  9,  9,  7, 10 },
	 {  8,  0, 11,  0,  7, 10 },
	 {  8,  0, 11, 12, 12, 10 }},
	{{  2,  0,  3,  3,  4,  0 }, //8
	 {  2,  5,  5,  5,  4,  0 },
	 {  1,  1,  6,  7,  8,  0 },
	 {  9,  9,  6,  7,  8,  0 },
	 {  0,  0,  0, 10, 11, 11 },
	 {  0,  0,  0, 10,  0,  0 }}
};
